/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#ifndef BIT_FIELDVECTOR_H
#define BIT_FIELDVECTOR_H

#include <string>
#include <vector>

#include <libxml++/libxml++.h>

#include <bit/field.h>

namespace bit
  {

  /**
   * A Field identifies a starting point and a length.
   *
   * This class extends FieldBase by including a name and a description.
   *
   * @ingroup records
   * @author Rick L. Vinyard, Jr.
   */
  class FieldVector: public Field
    {
    public:

      /**
      * Element is an base element in a field vector
      *
      * @author Rick L. Vinyard, Jr.
          */
    class Element: public FieldBase
        {
        public:

          typedef boost::shared_ptr<Element> pointer;

        protected:

          friend class FieldVector;

          Element(size_t index);

          static Element::pointer create(size_t index);

        public:

          virtual ~Element() { }

          /** @return the length of this object in it's length units */
          virtual size_t length() const;

          virtual size_t length(size_t units) const;

          /** @return the length units of this object */
          virtual size_t length_units() const;

          /** @return the offset of this object in the offset units of this object */
          virtual int offset() const;

          virtual int offset(size_t units) const;

          /**
           * The offset units of an element are always specified in the length units of elements.
           * In this way the offset is always specified as a multiple of length.
           */
          virtual size_t offset_units() const;

          /**
           * Returns the string name associated with this object
           *
           * This pure virtual method is reimplemented by children to
           * provide the proper notation, such as the array notation of
           * elements, by children.
           */
          virtual std::string name() const;

          virtual std::string name(int depth) const;

          /**
           * Returns the string description of this object.
           */
          virtual std::string description() const;

          size_t index() const;

          virtual FieldBase::pointer clone();

          protected:
          size_t m_index;
        };

      typedef boost::shared_ptr<FieldVector> pointer;

      /**
       * Contains a vector of fields.
       *
       * This class is provided to allow for efficient implementation of large arrays without allocating a field instance for
       * each element of the vector.
       *
       * @param name The base name of this vector
       * @param size The number of elements of this vector
       * @param element_length The length of each element in units
       * @param offset The offset of the vector only, not of each element. After the initial offset the vector is considered to be packed.
       * @param units The units of each element
       */
      FieldVector( std::string name="",
             size_t size=0,
             size_t element_length=0,
             size_t element_length_units=BITS,
             int offset=0,
             size_t offset_units=BITS,
             std::string descr=""
           );

      static FieldVector::pointer create( std::string name="",
                                    size_t size=0,
                                    size_t element_length=0,
                                    size_t element_length_units=BITS,
                                    int offset=0,
                                    size_t offset_units=BITS,
                                    std::string descr=""
                                  );

      virtual ~FieldVector();

      //       size_t read(xmlpp::Element& xml_element, size_t last_position);

      size_t length() const;

      size_t length(size_t units) const;

      size_t element_length() const;

      size_t size() const;

      void set_size(size_t size);

      virtual iterator begin();

      virtual iterator end();

      virtual FieldBase::pointer field(size_t i);

      virtual FieldBase::pointer clone();

    protected:
      size_t m_size;

      typedef std::vector<Element::pointer> Elements;
      
      Elements m_elements;

      virtual FieldBase::pointer previous_field(FieldBase::pointer current_field) throw (error::invalid_container_op);

      virtual FieldBase::pointer next_field(FieldBase::pointer current_field) throw (error::invalid_container_op);

    };

}

#endif
