/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#ifndef BITRECORD_H
#define BITRECORD_H

#include <string>

#include <libxml++/libxml++.h>

#include <bit/recordbase.h>
#include <bit/recordstorage.h>

/**
 * @defgroup records
 * Records and Fields are used to describe a parse tree to be applied to a linear data buffer.
 */

namespace bit
  {

    /**
     * bit::Record is a tree of bit::Field and bit::Record types sorted by
     * their starting bit positions.
     *
     * FOLDOC defines @b record as:
     * "An ordered set of fields, usually stored contiguously. ...
     * Some programming languages use the term to mean a type composed of fields of
     * several other types (C calls this a "struct")."
     *
     * In the context of the bit library, a record stores an ordered set of bit::Field types.
     *
     * The Record class extends the RecordBase by providing storage for Field objects as
     * well as providing the concept of a name and description strings.
     *
     * @ingroup records
     * @author Rick L. Vinyard, Jr.
     */
  class Record: public RecordBase
    {
    public:

      /** Convenience typedef to a smart pointer to bit::Record */
      typedef boost::shared_ptr<Record> pointer;

      /** Default constructor */
      Record( std::string name="",
              size_t length_units=BITS,
              int offset=0,
              size_t offset_units=BITS,
              std::string descr=""
            );

      /** Static method returns a smart pointer to a new Record object */
      static Record::pointer create( std::string name="",
                                     size_t length_units=BITS,
                                     int offset=0,
                                     size_t offset_units=BITS,
                                     std::string descr=""
                                   );

      /** Destructor */
      virtual ~Record();

      /** @return the length of this object in it's length units */
      virtual size_t length() const;

      virtual size_t length(size_t units) const;

      virtual iterator begin();

      virtual iterator end();

     virtual FieldBase::pointer field(std::string s);

      virtual FieldBase::pointer field(size_t i);

      virtual void add_field(FieldBase::pointer field);

      virtual void remove_field(FieldBase::pointer field);

      virtual void remove_field(const std::string& name);

      virtual size_t fields();

      virtual FieldBase::pointer clone();

    protected:

      RecordStorage m_fields;

      virtual FieldBase::pointer previous_field(FieldBase::pointer current_field) throw (error::invalid_container_op);

      virtual FieldBase::pointer next_field(FieldBase::pointer current_field) throw (error::invalid_container_op);
    };

}

#endif
