/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation version 2.1.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU Lesser General Public      *
 *   License along with this library; if not, write to the                 *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
 ***************************************************************************/
#ifndef BITRECORDBUFFER_H
#define BITRECORDBUFFER_H

#include <string>
#include <queue>
#include <set>

#include <bit/buffer.h>
#include <bit/record.h>
#include <bit/fieldbuffer.h>
#include <bit/data.h>
#include <bit/error.h>

namespace bit
  {

  /**
   * Combines the concepts of a Record or data layout with actual data.
   *
   * @ingroup buffers
   * @author Rick L. Vinyard, Jr.
   */
  class RecordBuffer : public Buffer
    {
    public:
      friend class FieldBuffer;

      RecordBuffer(size_t initial_size=0,
                   Growth growth=DYNAMIC,
                   size_t sizemax=0
                  );

      RecordBuffer(RecordBase::pointer record,
                   size_t initial_size=0,
                   Growth growth=DYNAMIC,
                   size_t sizemax=0
                  );

      RecordBuffer(void* external_data,
                   size_t data_octets,
                   DataMode datamode=COPY,
                   Growth growth=DYNAMIC,
                   size_t sizemax=0
                  );

      RecordBuffer(RecordBase::pointer record,
                   void* external_data,
                   size_t data_octets,
                   DataMode datamode=COPY,
                   Growth growth=DYNAMIC,
                   size_t sizemax=0
                  );

      virtual ~RecordBuffer();

      FieldBuffer operator[](size_t) throw (error::no_record);
      FieldBuffer operator[](std::string) throw (error::no_record);
      FieldBuffer operator[](FieldBase::iterator);
      FieldBuffer operator[](FieldBase::pointer);

      virtual void set_data(void* data, size_t size, DataMode datamode=COPY);

      void set_record(RecordBase::pointer record) throw ();
      void clear_record() throw ();
      RecordBase::pointer record();

      RecordBuffer& operator=(const RecordBuffer& other);

      sigc::signal<void, FieldBuffer> signal_field_changed();

      sigc::signal<void> signal_record_changed();

    protected:
      /**
       * A pointer to the record associated with this buffer, or NULL
       * if there is no record currently associated.
       */
      RecordBase::pointer m_record;

      /**
       * Causes the internal signal to be emitted notifying connected slots
       * that the data at field i has been changed.
       *
       * This method provides proper calling of the emit signal even when
       * a callback triggers a change. This would result in a recursive signal
       * which sigc++ doesn't allow.
       */
      void on_field_changed(FieldBase::pointer field);

      void on_record_changed();

    private:

      sigc::signal<void> m_signal_record_changed;

      /**
       * The signal emitted when an fieldsed value is changed.
       *
       * This member has been move to private, since derived children should not
       * directly emit the signal, but instead should call the notify_of_changes_to_fields
       * method.
       *
       * This method properly ensures that a signal is emitted for each changed fields,
       * even when a callback triggers an fields to change. This is necessary since
       * sigc++ does not allow recursive signals anymore.
       */
      sigc::signal<void, FieldBuffer> m_signal_field_changed;

      // Stores the fields that have changed and still need to have an emit signal
      //called. The fields are stored in the order they are changed.
      std::queue< FieldBase::pointer > m_signal_field_changed_queue;

      //Stores the set of fields that have changed and still need an emit signal.
      //The set is used to provide quick lookup O(lg n) rather than the O(n) necessary
      //in the queue.
      std::set< FieldBase::pointer >  m_signal_field_changed_set;

      bool  m_signal_field_changed_emitting;
    };

}

#endif
