/***************************************************************************
*   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
*   rvinyard@cs.nmsu.edu                                                  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation version 2.1.                *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU Lesser General Public      *
*   License along with this library; if not, write to the                 *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
***************************************************************************/
#ifndef BITUTILITY_H
#define BITUTILITY_H

#include <sys/types.h>
#include <stdint.h>
#include <string>

/** \file utility.h
 * This header contains various utility functions inside the bit namespace.
 *
 * @defgroup utility
 * Many of these functions are meant to extend the concept of htonl, htons, ntohl
 * and ntohs. Several decisions were made when creating these functions:
 *  - Using functions instead of macros allows:
 *    - similar performance gains when using a decent optimizing compiler.
 *    - typechecking on arguments.
 *    - constraining functions to the bit namespace.
 *  - Rather than just converting host <==> network byte orders, there are explict
 *    functions for converting:
 *    - host <==> little-endian
 *    - host <==> big-endian
 *  - The following notations are used:
 *    - \c host instead of \c h
 *      - Personally, I don't think the extra \c ost is too much more to type and I think it
 *        adds to readability
 *    - \c net instead of \c n for the same reasons as \c host
 *    - \c le for little-endian
 *      - Here, I thought there was a difference between \c le and \c littleendian. Somehow the
 *        latter seems awkward.
 *    - \c be for big-endian
 *  - Additionally, there are byte swapping functions for \c floats and \c doubles as well as
 *    16, 32 and 64-bit integers
 */

namespace bit {

/** convert 64 bit value x from big-endian to host byte order
 * @ingroup utility
 */
uint64_t be_to_host( uint64_t x);

/** convert 64 bit value x from big-endian to host byte order
 * @ingroup utility
 */
int64_t be_to_host( int64_t x);

/** convert 32 bit value x from big-endian to host byte order
 * @ingroup utility
 */
uint32_t be_to_host( uint32_t x);

/** convert 32 bit value x from big-endian to host byte order
 * @ingroup utility
 */
int32_t be_to_host( int32_t x);

/** convert 16 bit value x from big-endian to host byte order
 * @ingroup utility
 */
uint16_t be_to_host( uint16_t x);

/** convert 16 bit value x from big-endian to host byte order
 * @ingroup utility
 */
int16_t be_to_host( int16_t x);

/** convert 64 bit value x from host to big-endian byte order
 * @ingroup utility
 */
uint64_t host_to_be( uint64_t x);

/** convert 64 bit value x from host to big-endian byte order
 * @ingroup utility
 */
int64_t host_to_be( int64_t x);

/** convert 32 bit value x from host to big-endian byte order
 * @ingroup utility
 */
uint32_t host_to_be( uint32_t x);

/** convert 32 bit value x from host to big-endian byte order
 * @ingroup utility
 */
int32_t host_to_be( int32_t x);

/** convert 16 bit value x from host to big-endian byte order
 * @ingroup utility
 */
uint16_t host_to_be( uint16_t x);

/** convert 16 bit value x from host to big-endian byte order
 * @ingroup utility
 */
int16_t host_to_be( int16_t x);

/** convert 64 bit value x from little-endian to host byte order
 * @ingroup utility
 */
uint64_t le_to_host( uint64_t x);

/** convert 64 bit value x from little-endian to host byte order
 * @ingroup utility
 */
int64_t le_to_host( int64_t x);

/** convert 32 bit value x from little-endian to host byte order
 * @ingroup utility
 */
uint32_t le_to_host( uint32_t x);

/** convert 32 bit value x from little-endian to host byte order
 * @ingroup utility
 */
int32_t le_to_host( int32_t x);

/** convert 16 bit value x from little-endian to host byte order
 * @ingroup utility
 */
uint16_t le_to_host( uint16_t x);

/** convert 16 bit value x from little-endian to host byte order
 * @ingroup utility
 */
int16_t le_to_host( int16_t x);

/** convert 64 bit value x from host to little-endian byte order
 * @ingroup utility
 */
uint64_t host_to_le( uint64_t x);

/** convert 64 bit value x from host to little-endian byte order
 * @ingroup utility
 */
int64_t host_to_le( int64_t x);

/** convert 32 bit value x from host to little-endian byte order
 * @ingroup utility
 */
uint32_t host_to_le( uint32_t x);

/** convert 32 bit value x from host to little-endian byte order
 * @ingroup utility
 */
int32_t host_to_le( int32_t x);

/** convert 16 bit value x from host to little-endian byte order
 * @ingroup utility
 */
uint16_t host_to_le( uint16_t x);

/** convert 16 bit value x from host to little-endian byte order
 * @ingroup utility
 */
int16_t host_to_le( int16_t x);

/** convert 64 bit value x from network to host byte order
 * @ingroup utility
 */
uint64_t net_to_host( uint64_t x);

/** convert 64 bit value x from network to host byte order
 * @ingroup utility
 */
int64_t net_to_host( int64_t x);

/** convert 32 bit value x from network to host byte order
 * @ingroup utility
 */
uint32_t net_to_host( uint32_t x);

/** convert 32 bit value x from network to host byte order
 * @ingroup utility
 */
int32_t net_to_host( int32_t x);

/** convert 16 bit value x from network to host byte order
 * @ingroup utility
 */
uint16_t net_to_host( uint16_t x);

/** convert 16 bit value x from network to host byte order
 * @ingroup utility
 */
int16_t net_to_host( int16_t x);

/** convert 64 bit value x from host to network byte order
 * @ingroup utility
 */
uint64_t host_to_net( uint64_t x);

/** convert 64 bit value x from host to network byte order
 * @ingroup utility
 */
int64_t host_to_net( int64_t x);

/** convert 32 bit value x from host to network byte order
 * @ingroup utility
 */
uint32_t host_to_net( uint32_t x);

/** convert 32 bit value x from host to network byte order
 * @ingroup utility
 */
int32_t host_to_net( int32_t x);

/** convert 16 bit value x from host to network byte order
 * @ingroup utility
 */
uint16_t host_to_net( uint16_t x);

/** convert 16 bit value x from host to network byte order
 * @ingroup utility
 */
int16_t host_to_net( int16_t x);

/** convert double value x from network to host byte order
 * @ingroup utility
 */
double be_to_host( double x);

/** convert double value x from network to host byte order
 * @ingroup utility
 */
double le_to_host( double x);

/** convert double value x from network to host byte order
 * @ingroup utility
 */
double net_to_host( double x);

/** convert float value x from network to host byte order
 * @ingroup utility
 */
float be_to_host( float x);

/** convert float value x from network to host byte order
 * @ingroup utility
 */
float le_to_host( float x);

/** convert float value x from network to host byte order
 * @ingroup utility
 */
float net_to_host( float x);

/** convert double value x from host to network byte order
 * @ingroup utility
 */
double host_to_be( double x);

/** convert double value x from host to network byte order
 * @ingroup utility
 */
double host_to_le( double x);

/** convert double value x from host to network byte order
 * @ingroup utility
 */
double host_to_net( double x);

/** convert float value x from host to network byte order
 * @ingroup utility
 */
float host_to_be( float x);

/** convert float value x from host to network byte order
 * @ingroup utility
 */
float host_to_le( float x);

/** convert float value x from host to network byte order
 * @ingroup utility
 */
float host_to_net( float x);


/* Modified from GLIB gtypes.h */

/* GLIB - Library of useful routines for C programming
 * Copyright (C) 1995-1997  Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the GLib Team and others 1997-2000.  See the AUTHORS
 * file for a list of people on the GLib Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GLib at ftp://ftp.gtk.org/pub/gtk/.
 */



/* IEEE Standard 754 Single Precision Storage Format (gfloat):
 *
 *        31 30           23 22            0
 * +--------+---------------+---------------+
 * | s 1bit | e[30:23] 8bit | f[22:0] 23bit |
 * +--------+---------------+---------------+
 * B0------------------->B1------->B2-->B3-->
 *
 * IEEE Standard 754 Double Precision Storage Format (gdouble):
 *
 *        63 62            52 51            32   31            0
 * +--------+----------------+----------------+ +---------------+
 * | s 1bit | e[62:52] 11bit | f[51:32] 20bit | | f[31:0] 32bit |
 * +--------+----------------+----------------+ +---------------+
 * B0--------------->B1---------->B2--->B3---->  B4->B5->B6->B7->
 */
/* subtract from biased_exponent to form base2 exponent (normal numbers) */
#define IEEE754_FLOAT_BIAS  (127)
#define IEEE754_DOUBLE_BIAS (1023)
/* multiply with base2 exponent to get base10 exponent (normal numbers) */
#define LOG_2_BASE_10   (0.30102999566398119521)

union lefloatIEEE754
{
  float f;
  struct {
    unsigned int mantissa : 23;
    unsigned int biased_exponent : 8;
    unsigned int sign : 1;
  } mpn;
};

union ledoubleIEEE754
{
  double d;
  struct {
    unsigned int mantissa_low : 32;
    unsigned int mantissa_high : 20;
    unsigned int biased_exponent : 11;
    unsigned int sign : 1;
  } mpn;
};

union befloatIEEE754
{
  float f;
  struct {
    unsigned int sign : 1;
    unsigned int biased_exponent : 8;
    unsigned int mantissa : 23;
  } mpn;
};

union bedoubleIEEE754
{
  double d;
  struct {
    unsigned int sign : 1;
    unsigned int biased_exponent : 11;
    unsigned int mantissa_high : 20;
    unsigned int mantissa_low : 32;
  } mpn;
};

/**
 * Calculate the number of octets required to completely hold n bits.
 * ex: octet_ceiling(12) returns 2.
 */
size_t octet_ceiling(size_t bits);

/**
 * Calculate the number of whole octets required to hold n bits.
 * ex: octet_floor(12) returns 1.
 */
size_t octet_floor(size_t bits);

/**
 * Given a pointer p to a memory segment and a bit offset, return
 * a pointer to the octet that contains the start of the sequence.
 */
void* starting_octet(void* p, size_t bits);

/**
 * Given a pointer p to a memory segment and a bit offset, return
 * a pointer to the octet that contains the end of the sequence.
 */
void* ending_octet(void* p, size_t bits);

/**
 * Given a bit sequence starting at zero, calculate the number of
 * of upper (used) bits in the last octet.
 * ex: last_octet_upper_bits(10) returns 2.
 */
size_t last_octet_upper_bits(size_t bits);

/**
 * Given a bit sequence starting at zero, calculate the number of
 * lower (unused) bits in the last octet.
 * ex: last_octet_lower_bits(10) returns 6.
 */
size_t last_octet_lower_bits(size_t bits);

/** Left shift an entire memory segment by a given number of bits. */
void* left_shift(void* buffer, size_t buffer_size, size_t lshift_bits);

/** Right shift an enitre memory segment by a given number of bits */
void* right_shift(void* buffer, size_t buffer_size, size_t rshift_bits);

std::string hex_string( void* buffer, size_t buffer_size,
                        bool uppercase=true,
                        std::string separator="",
                        std::string prefix="0x",
                        std::string postfix=""
                      );

std::string binary_string( void* buffer, size_t buffer_size,
                           std::string separator="",
                           size_t separator_digits=8,
                           std::string prefix="",
                           std::string postfix=""
                         );

} // namespace bit

#endif

